import pytest
import spacy
from tempfile import TemporaryDirectory
from thinc.api import Ragged, get_current_ops

from curated_transformers.tokenization.hf_loader import build_hf_piece_encoder_loader_v1
from curated_transformers.tokenization.wordpiece_encoder import (
    build_bert_wordpiece_encoder_v1,
    build_wordpiece_encoder_v1,
    _bert_preprocess,
    build_wordpiece_encoder_loader_v1,
)
from curated_transformers._compat import has_hf_transformers, transformers


def test_wordpiece_encoder_local_model(wordpiece_toy_encoder, sample_docs):
    encoding = wordpiece_toy_encoder.predict(sample_docs)
    _check_toy_encoder(encoding)


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_wordpiece_encoder_hf_model(sample_docs):
    ops = get_current_ops()
    encoder = build_wordpiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="bert-base-cased")
    encoder.initialize()

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [101, 146, 1486, 170, 1873, 1114, 170, 16737, 119, 102]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 1, 1, 1, 1, 3, 1, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd,
        [101, 3570, 1195, 1209, 3940, 185, 5926, 2744, 7329, 119, 102],
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_wordpiece_encoder_hf_model_uncased(sample_docs):
    ops = get_current_ops()
    encoder = build_wordpiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="bert-base-uncased")
    encoder.initialize()

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [101, 1045, 2387, 1037, 2611, 2007, 1037, 12772, 1012, 102]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 1, 1, 1, 1, 1, 1, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd, [101, 2651, 2057, 2097, 4521, 26202, 4605, 1012, 102]
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_wordpiece_encoder_hf_model_german():
    ops = get_current_ops()
    encoder = build_bert_wordpiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="bert-base-german-cased")
    encoder.initialize()

    nlp = spacy.blank("de")
    sample_docs = [
        nlp.make_doc("Wir sehen ein AWO-Mitarbeiter."),
        nlp.make_doc("Die Mw.-St. beträgt 19 Prozent."),
    ]

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(encoding[0].lengths, [1, 1, 1, 1, 5, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [3, 655, 2265, 39, 32, 26939, 26962, 26935, 2153, 26914, 4]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 1, 5, 1, 1, 1, 1, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd,
        [3, 125, 56, 26915, 26914, 26935, 130, 26914, 4490, 141, 1028, 26914, 4],
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_wordpiece_encoder_loader(sample_docs):
    ops = get_current_ops()
    encoder = build_wordpiece_encoder_v1()
    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("bert-base-cased")
    with TemporaryDirectory() as d:
        vocab_path = hf_tokenizer.save_vocabulary(d)[0]
        encoder.init = build_wordpiece_encoder_loader_v1(path=vocab_path)
        encoder.initialize()

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [101, 146, 1486, 170, 1873, 1114, 170, 16737, 119, 102]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 1, 1, 1, 1, 3, 1, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd,
        [101, 3570, 1195, 1209, 3940, 185, 5926, 2744, 7329, 119, 102],
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_wordpiece_encoder_loader_uncased(sample_docs):
    ops = get_current_ops()
    encoder = build_wordpiece_encoder_v1()
    hf_tokenizer = transformers.AutoTokenizer.from_pretrained("bert-base-uncased")
    with TemporaryDirectory() as d:
        vocab_path = hf_tokenizer.save_vocabulary(d)[0]
        encoder.init = build_wordpiece_encoder_loader_v1(
            path=vocab_path, lowercase=True, strip_accents=True
        )
        encoder.initialize()

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [101, 1045, 2387, 1037, 2611, 2007, 1037, 12772, 1012, 102]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 1, 1, 1, 1, 1, 1, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd, [101, 2651, 2057, 2097, 4521, 26202, 4605, 1012, 102]
    )


def test_serialize(wordpiece_toy_encoder):
    encoder_bytes = wordpiece_toy_encoder.to_bytes()
    toy_encoder2 = build_wordpiece_encoder_v1()
    toy_encoder2.from_bytes(encoder_bytes)
    assert (
        wordpiece_toy_encoder.attrs["wordpiece_processor"].to_list()
        == toy_encoder2.attrs["wordpiece_processor"].to_list()
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_serialize_hf_model():
    encoder = build_wordpiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="bert-base-cased")
    encoder.initialize()
    encoder_bytes = encoder.to_bytes()
    encoder2 = build_wordpiece_encoder_v1()
    encoder2.from_bytes(encoder_bytes)
    assert (
        encoder.attrs["wordpiece_processor"].to_list()
        == encoder2.attrs["wordpiece_processor"].to_list()
    )


def test_bert_preprocess():
    assert _bert_preprocess("AWO-Mitarbeiter") == ["AWO", "-", "Mitarbeiter"]
    assert _bert_preprocess("-Mitarbeiter") == ["-", "Mitarbeiter"]
    assert _bert_preprocess("AWO-") == ["AWO", "-"]
    assert _bert_preprocess("-") == ["-"]
    assert _bert_preprocess("") == []
    assert _bert_preprocess("Mw.-St.") == ["Mw", ".", "-", "St", "."]


def _check_toy_encoder(encoding):
    ops = get_current_ops()
    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 3, 1, 1, 5, 1, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd,
        [2, 41, 818, 61, 67, 193, 88, 204, 61, 251, 909, 682, 102, 95, 17, 3],
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 3, 1, 1, 2, 3, 3, 1, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd,
        [2, 824, 98, 189, 311, 417, 65, 155, 503, 99, 1, 416, 117, 88, 17, 3],
    )



from typing import Any, Callable, List, TypeVar

from spacy.tokens.doc import Doc
from spacy.tokens.token import Token
from thinc.model import Model
from thinc.types import Ragged


TokenSequenceT = TypeVar("TokenSequenceT", Doc, List[Token])

Tok2PiecesInT = List[TokenSequenceT]
Tok2PiecesOutT = List[Ragged]
Tok2PiecesBackpropT = Callable[[Any], Any]
Tok2PiecesModelT = Model[Tok2PiecesInT, Tok2PiecesOutT]

PieceAdapterInOutT = List[Ragged]
PieceAdapterBackpropT = Callable[[Any], Any]
PieceAdapterModelT = Model[PieceAdapterInOutT, PieceAdapterInOutT]



from .architectures import (
    build_albert_transformer_model_v1,
    build_bert_transformer_model_v1,
    build_camembert_transformer_model_v1,
    build_roberta_transformer_model_v1,
    build_xlmr_transformer_model_v1,
    build_pytorch_checkpoint_loader_v1,
)
from .hf_loader import build_hf_transformer_encoder_loader_v1
from .scalar_weight import build_scalar_weight_v1
from .with_strided_spans import build_with_strided_spans_v1



import pytest
from thinc.api import Ragged, registry, get_current_ops

from curated_transformers.tokenization.sentencepiece_encoder import (
    build_sentencepiece_encoder_v1,
)
from curated_transformers.tokenization.hf_loader import build_hf_piece_encoder_loader_v1
from curated_transformers._compat import has_hf_transformers


@pytest.fixture
def toy_model_path(test_dir):
    return test_dir / "toy.model"


@pytest.fixture
def toy_encoder(toy_model_path):
    encoder = build_sentencepiece_encoder_v1()
    encoder.init = registry.model_loaders.get(
        "curated-transformers.SentencepieceLoader.v1"
    )(path=toy_model_path)
    encoder.initialize()
    return encoder


def test_sentencepiece_encoder(toy_encoder, sample_docs):
    encoding = toy_encoder.predict(sample_docs)
    _check_toy_encoder(encoding)


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_sentencepiece_encoder_hf_model(sample_docs):
    encoder = build_sentencepiece_encoder_v1()
    encoder.init = build_hf_piece_encoder_loader_v1(name="xlm-roberta-base")
    encoder.initialize()

    encoding = encoder.predict(sample_docs)

    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops = get_current_ops()
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 2, 2, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd, [1, 86, 24123, 9, 23039, 677, 9, 5500, 70819, 5, 4, 2]
    )

    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 1, 1, 1, 1, 2, 1, 2, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd, [1, 38395, 641, 1220, 73202, 159, 7663, 120323, 5, 4, 2]
    )


def test_serialize(toy_encoder, sample_docs):
    encoder_bytes = toy_encoder.to_bytes()
    encoder2 = build_sentencepiece_encoder_v1()
    encoder2.from_bytes(encoder_bytes)
    encoding = encoder2.predict(sample_docs)
    _check_toy_encoder(encoding)


def _check_toy_encoder(encoding):
    assert isinstance(encoding, list)
    assert len(encoding) == 2

    assert isinstance(encoding[0], Ragged)
    ops = get_current_ops()
    ops.xp.testing.assert_array_equal(
        encoding[0].lengths, [1, 1, 1, 1, 1, 1, 1, 6, 2, 1]
    )
    ops.xp.testing.assert_array_equal(
        encoding[0].dataXd,
        [1, 8, 465, 10, 947, 41, 10, 170, 168, 110, 28, 20, 143, 7, 4, 2],
    )

    assert isinstance(encoding[1], Ragged)
    ops.xp.testing.assert_array_equal(encoding[1].lengths, [1, 2, 1, 1, 1, 4, 3, 2, 1])
    ops.xp.testing.assert_array_equal(
        encoding[1].dataXd,
        [1, 483, 546, 112, 171, 567, 62, 20, 45, 0, 84, 115, 27, 7, 4, 2],
    )



from typing import List
from curated_transformers.models.output import TransformerModelOutput
import pytest
from thinc.api import Model, get_current_ops, Ragged, with_array, chain
from thinc.types import Floats2d

from curated_transformers.models.with_strided_spans import with_strided_spans


def relu_activation() -> Model[Floats2d, Floats2d]:
    def forward(model: Model, X: Floats2d, is_train: bool):
        Y = model.ops.relu(X)

        def backprop(dY: Floats2d) -> Floats2d:
            return dY * model.ops.backprop_relu(dY, Y)

        return Y, backprop

    return Model("relu_activation", forward)


def _add_range() -> Model[Floats2d, Floats2d]:
    """Add range [0, X.size)."""

    def forward(model: Model, X: Floats2d, is_train: bool):
        adds = model.ops.xp.arange(X.size).reshape(X.shape)
        return X + adds, lambda x: x

    return Model("add_range", forward)


def _mock_transformer() -> Model[List[Floats2d], TransformerModelOutput]:
    def forward(model: Model, X: List[Floats2d], is_train: bool):
        return (
            TransformerModelOutput(outputs=[[x] for x in X], last_layer_only=True),
            lambda x: x,
        )

    return Model("mock_transformer", forward)


@pytest.mark.parametrize("batch_size", [1, 2, 3, 5])
def test_with_strided_spans(batch_size):
    ops = get_current_ops()
    trf = chain(with_array(relu_activation()), _mock_transformer())
    model = with_strided_spans(trf, stride=4, window=4, batch_size=batch_size)

    zeros = ops.alloc2f(15, 5)
    ones = ops.alloc2f(15, 5) + 1
    fives = ops.alloc2f(15, 5) + 5

    lengths1 = ops.asarray1i([1, 2, 3, 4, 5])
    lengths2 = ops.asarray1i([5, 4, 3, 2, 1])

    X = [
        Ragged(fives.copy(), lengths=lengths1),
        Ragged(-fives, lengths=lengths2),
        Ragged(fives.copy(), lengths=lengths1.copy()),
        Ragged(-fives, lengths=lengths2.copy()),
    ]
    model.initialize(X)

    Y, backprop = model(X, is_train=True)
    Y = Y.all_outputs
    assert len(Y) == len(X)

    dX = backprop(
        [
            Ragged(ones.copy(), lengths=lengths1),
            Ragged(ones.copy(), lengths=lengths2),
            Ragged(ones.copy(), lengths=lengths1.copy()),
            Ragged(ones.copy(), lengths=lengths2.copy()),
        ]
    )

    for i in [0, 2]:
        ops.xp.testing.assert_array_equal(Y[i][0].data, fives)
        ops.xp.testing.assert_array_equal(Y[i][0].lengths, lengths1)
        ops.xp.testing.assert_array_equal(dX[i].data, ones)
        ops.xp.testing.assert_array_equal(dX[i].lengths, lengths1)
    for i in [1, 3]:
        ops.xp.testing.assert_array_equal(Y[i][0].data, zeros)
        ops.xp.testing.assert_array_equal(Y[i][0].lengths, lengths2)
        ops.xp.testing.assert_array_equal(dX[i].data, zeros)
        ops.xp.testing.assert_array_equal(dX[i].lengths, lengths2)


def test_with_strided_spans_averaging():
    ops = get_current_ops()
    stateful = chain(with_array(_add_range()), _mock_transformer())
    model = with_strided_spans(stateful, stride=2, window=4)

    data = ops.xp.zeros((6, 2))
    lengths = ops.asarray1i([3, 3])
    X = [Ragged(data, lengths=lengths)]

    model.initialize(X)

    Y, backprop = model(X, is_train=False)

    ops.xp.testing.assert_array_equal(
        Y.all_outputs[0][0].dataXd,
        [[0.0, 1.0], [2.0, 3.0], [6.0, 7.0], [8.0, 9.0], [14.0, 15.0], [16.0, 17.0]],
    )

    ones = data + 1
    dX = backprop(
        [
            Ragged(ones.copy(), lengths=lengths),
        ]
    )
    ops.xp.testing.assert_array_equal(
        dX[0].dataXd,
        [
            [1.0, 1.0],
            [1.0, 1.0],
            [0.25, 0.25],
            [0.25, 0.25],
            [0.25, 0.25],
            [0.25, 0.25],
        ],
    )
    ops.xp.testing.assert_array_equal(dX[0].lengths, lengths)


def test_incorrect_strides_are_rejected():
    relu = with_array(relu_activation())
    with pytest.raises(ValueError):
        with_strided_spans(relu, stride=2, window=6)
    with pytest.raises(ValueError):
        with_strided_spans(relu, stride=4, window=3)


def test_batch_sizes_are_rejected():
    relu = with_array(relu_activation())
    with pytest.raises(ValueError):
        with_strided_spans(relu, batch_size=-1)
    with pytest.raises(ValueError):
        with_strided_spans(relu, batch_size=0)



from typing import Any, Callable, Iterable, List, Optional, Tuple, Union, cast
from functools import partial
from thinc.model import Model
from thinc.types import Ragged, Floats2d, Ints1d

from .output import TransformerModelOutput
from .types import (
    RaggedInOutT,
    Floats2dInOutT,
    SpanExtractorBackpropT,
    SpanExtractorInT,
    SpanExtractorOutT,
    SpanExtractorModelT,
    TorchTransformerInT,
    TorchTransformerModelT,
    TorchTransformerOutT,
)
from ..errors import Errors


def build_with_strided_spans_v1(
    stride: int = 96, window: int = 128, batch_size: int = 384
) -> Callable[[TorchTransformerModelT, int, int], SpanExtractorModelT]:
    """Construct a model that can be used to convert a list of ragged
    piece identifiers to strided spans and vice-versa.

    stride (int):
        The stride of the span generator.
    window (int):
        The (maximum) size of each span.
    batch_size (int):
        The maximum number of spans that are processed
        by the transformer model at a time.
    """
    return partial(
        with_strided_spans, stride=stride, window=window, batch_size=batch_size
    )


def with_strided_spans(
    trf_model: TorchTransformerModelT,
    *,
    stride: int = 96,
    window: int = 128,
    batch_size: int = 384,
) -> Model[List[Ragged], TransformerModelOutput]:
    if not (window // 2 <= stride <= window):
        raise ValueError(
            Errors.E017.format(
                stride=stride, half_window_size=window // 2, window_size=window
            )
        )
    if batch_size <= 0:
        raise ValueError(Errors.E018)

    attrs = {
        "stride": stride,
        "window": window,
        "batch_size": batch_size,
    }
    return Model(
        "with_strided_spans",
        with_strided_spans_forward,
        init=with_strided_spans_init,
        layers=[trf_model],
        attrs=attrs,
    )


def with_strided_spans_init(
    model: SpanExtractorModelT,
    X: Optional[SpanExtractorInT],
    Y: Optional[SpanExtractorInT],
):
    stride: int = model.attrs["stride"]
    window: int = model.attrs["window"]

    X_spans = (
        _ragged_to_strided_arrays(X, stride=stride, window=window)[0]
        if X is not None
        else None
    )
    Y_spans = (
        _ragged_to_strided_arrays(Y, stride=stride, window=window)[0]
        if Y is not None
        else None
    )
    model.layers[0].initialize(X=X_spans, Y=Y_spans)


def with_strided_spans_forward(
    model: SpanExtractorModelT,
    X: SpanExtractorInT,
    is_train: bool,
) -> Tuple[SpanExtractorOutT, SpanExtractorBackpropT]:
    transformer: TorchTransformerModelT = model.layers[0]
    stride: int = model.attrs["stride"]
    window: int = model.attrs["window"]
    batch_size: int = model.attrs["batch_size"]

    spans, doc_lens = _ragged_to_strided_arrays(Xlr=X, stride=stride, window=window)
    # Calculate once and use for all layer outputs.
    doc_len_sums = [int(x.sum()) for x in doc_lens]

    backprops = []
    outputs = []
    for batch in _split_spans(spans, batch_size):
        output, bp = transformer(cast(TorchTransformerInT, batch), is_train=is_train)
        if not isinstance(output, TransformerModelOutput):
            raise ValueError(
                Errors.E014.format(model_name=model.name, input_type=type(output))
            )
        outputs.append(output)
        backprops.append(bp)

    trf_output = _unsplit_outputs(outputs)

    # Suppose we have:
    #
    # A B C D E F G H I
    # <--->   stride
    # <-----> window
    #
    # D is part of both the current and the next window. To
    # ensure that there is no drift in representations, we
    # average them in both representations.
    _apply_to_overlaps(
        trf_output.all_outputs,
        doc_len_sums,
        stride=stride,
        window=window,
        func=_average_arrays,
    )

    model_output = _strided_arrays_to_ragged(
        model, trf_output.all_outputs, doc_lens, doc_len_sums, stride=stride
    )
    trf_output.all_outputs = cast(List[List[Ragged]], model_output)

    def backprop(dY: RaggedInOutT):
        dY_spans, dY_lengths = _ragged_to_strided_arrays(
            dY, stride=stride, window=window
        )
        # Calculate once and use for all layer outputs.
        dY_length_sums = [int(x.sum()) for x in dY_lengths]

        # Both overlaps will have dY. However, the proper gradient is 0.5 * dY
        # since we averaged the overlaps in the forward pass.
        _apply_to_overlaps(
            dY_spans,
            dY_length_sums,
            stride=stride,
            window=window,
            func=_normalize_gradients,
        )

        dXlf = []
        split_dY_spans = list(_split_spans(dY_spans, batch_size))
        assert len(split_dY_spans) == len(backprops)
        for dY_batch, bp_trf in zip(split_dY_spans, backprops):
            dXlf_batch = bp_trf(dY_batch)
            dXlf.extend(dXlf_batch)

        return _strided_arrays_to_ragged(
            model, dXlf, dY_lengths, dY_length_sums, stride=stride
        )

    return trf_output, backprop


def _split_spans(spans: Floats2dInOutT, batch_size: int) -> Iterable[Floats2dInOutT]:
    while len(spans):
        batch = spans[:batch_size]
        yield batch
        spans = spans[batch_size:]


def _unsplit_outputs(outputs: List[TorchTransformerOutT]) -> TorchTransformerOutT:
    last_layer_only = outputs[0].last_layer_only
    merged = TransformerModelOutput(
        outputs=cast(
            List[List[Floats2d]],
            [inner for outer in outputs for inner in outer.all_outputs],
        ),
        last_layer_only=last_layer_only,
    )
    return merged


def _apply_to_overlaps(
    Xlf: Floats2dInOutT,
    len_sums: List[int],
    *,
    stride: int,
    window: int,
    func: Callable[[Any, Any], None],
):
    """Applies a function to overlapping windows, modifying the arrays
    in Xlf in-place."""

    def _apply_to_layer(input: Union[Tuple[Floats2d, ...], List[Floats2d]]):
        for doc_len in len_sums:
            prev_overlap = None
            while doc_len != 0:
                overlap = min(window - stride, doc_len)
                if prev_overlap is not None:
                    cur_overlap = input[0][:overlap]
                    prev_overlap = prev_overlap[-overlap:]
                    func(prev_overlap, cur_overlap)
                prev_overlap = input[0][-overlap:]

                doc_len -= min(stride, doc_len)
                input = input[1:]

    def _apply_to_layers(input: List[List[Floats2d]]):
        # We need to transpose the input since the overlaps happen between
        # each layer of each span, i.e., span 1 layer 1 overlaps with span 2 layer 1, etc.
        transposed = list(zip(*input))
        for x in transposed:
            _apply_to_layer(x)

    # Nothing to do if there is no overlap.
    if window - stride == 0:
        return

    if isinstance(Xlf[0], list):
        nested_list = cast(List[List[Floats2d]], Xlf)
        _apply_to_layers(nested_list)
    else:
        flat_tuple = cast(Tuple[Floats2d, ...], Xlf)
        _apply_to_layer(flat_tuple)


def _average_arrays(array1, array2):
    array1 += array2
    array1 /= 2.0
    array2[:] = array1


def _normalize_gradients(array1, array2):
    array1 *= 0.5
    array2 *= 0.5


def _ragged_to_strided_arrays(
    Xlr: RaggedInOutT, *, stride: int, window: int
) -> Tuple[Floats2dInOutT, List[Ints1d]]:
    """Convert the per-Doc Ragged sequences to an array of sub-sequences that span
    all the input documents."""

    def _apply_to_layer(
        input: Union[Tuple[Ragged, ...], List[Ragged]], output: List[Floats2d]
    ):
        for Xr in input:
            data = cast(Floats2d, Xr.dataXd)
            while data.shape[0] != 0:
                output.append(data[:window])
                data = data[stride:]

    def _apply_to_layers(input: List[List[Ragged]]):
        # Transpose input to reconstruct strides across all documents.
        transposed = list(zip(*Xlr))
        spans: List[List[Floats2d]] = [[] for _ in range(len(transposed))]
        lens = [x.lengths for x in transposed[0]]
        for x, y in zip(transposed, spans):
            _apply_to_layer(x, y)

        # Normalize sequences as lists.
        spans = [[y for y in x] for x in zip(*spans)]
        return spans, lens

    if isinstance(Xlr[0], list):
        # Gradients in the backward pass.
        # Shape of spans: (span, layer)
        nested_list = cast(List[List[Ragged]], Xlr)
        return _apply_to_layers(nested_list)
    else:
        # Inputs in the forward pass.
        flat_list = cast(List[Ragged], Xlr)
        spans: List[Floats2d] = []
        lens = [x.lengths for x in flat_list]
        _apply_to_layer(flat_list, spans)
        return spans, lens


def _strided_arrays_to_ragged(
    model: Model,
    Xlf: Floats2dInOutT,
    lens: List[Ints1d],
    len_sums: List[int],
    *,
    stride: int,
) -> RaggedInOutT:
    """Inverse operation of _ragged_to_strided_arrays."""

    def _apply_to_layer(
        input: Union[Tuple[Floats2d, ...], List[Floats2d]], output: List[Ragged]
    ):
        for doc_len, Xr_lens in zip(len_sums, lens):
            arrs = []
            while doc_len != 0:
                arr = input[0][: min(stride, doc_len)]
                arrs.append(arr)
                doc_len -= arr.shape[0]
                input = input[1:]
            output.append(Ragged(model.ops.flatten(arrs), lengths=Xr_lens))

    def _apply_to_layers(input: List[List[Floats2d]]):
        # Transpose input to reconstruct strides across all documents.
        transposed = list(zip(*input))
        Xlr: List[List[Ragged]] = [[] for _ in range(len(transposed))]
        for x, y in zip(transposed, Xlr):
            _apply_to_layer(x, y)

        # Normalize sequences as lists.
        Xlr = [[y for y in x] for x in zip(*Xlr)]
        return Xlr

    if isinstance(Xlf[0], list):
        # Transformer output in forward pass.
        nested_list = cast(List[List[Floats2d]], Xlf)
        return _apply_to_layers(nested_list)
    else:
        # Gradient from torch in backward pass.
        flat_tuple = cast(Tuple[Floats2d, ...], Xlf)
        Xlr: List[Ragged] = []
        _apply_to_layer(flat_tuple, Xlr)
        return Xlr



from torch import Tensor
from torch.nn import Module, ModuleList

from ..attention import AttentionMask
from ..bert.config import BertAttentionConfig
from ..bert.layer import BertAttentionConfig, BertEncoderLayer
from .config import AlbertLayerConfig


class AlbertLayerGroup(Module):
    def __init__(
        self, layer_config: AlbertLayerConfig, attention_config: BertAttentionConfig
    ) -> None:
        super().__init__()

        self.group_layers = ModuleList(
            [
                BertEncoderLayer(layer_config, attention_config)
                for _ in range(layer_config.inner_group_num)
            ]
        )

    def forward(self, input: Tensor, attn_mask: AttentionMask) -> Tensor:
        """
        Shapes:
            input - (batch, seq_len, width)
        """
        layer_output = input
        for layer in self.group_layers:
            layer_output = layer(layer_output, attn_mask)
        return layer_output



from typing import Optional
import math
import torch
from torch import Tensor
from torch.nn import Module

from ...errors import Errors


class AttentionMask:
    bool_mask: Tensor
    _logit_mask: Optional[Tensor]

    def __init__(self, bool_mask: Tensor):
        if bool_mask.dtype != torch.bool:
            raise ValueError("Expected the attention mask to be of dtype 'torch.bool'")
        self.bool_mask = bool_mask
        self._logit_mask = torch.jit.annotate(Optional[Tensor], None)

    @property
    def logit_mask(self) -> Tensor:
        if self._logit_mask is None:
            # The value is `torch.finfo(attn_scores.dype).min`. Unfortunately,
            # we cannot use `torch.finfo` in TorchScript.
            self._logit_mask = (1.0 - self.bool_mask.int()) * -3.4028234663852886e38

        # Narrow type for TorchScript.
        logit_mask = self._logit_mask
        assert logit_mask is not None
        return logit_mask

    def dim(self) -> int:
        return self.bool_mask.dim()

    @property
    def shape(self):
        return self.bool_mask.shape


# https://www.tensorflow.org/text/tutorials/transformer#scaled_dot_product_attention
class ScaledDotProductAttention(Module):
    def __init__(self, *, dropout_prob: float = 0.1):
        super().__init__()
        self.dropout = torch.nn.Dropout(p=dropout_prob)

    def forward(
        self, k: Tensor, q: Tensor, v: Tensor, attn_mask: AttentionMask
    ) -> Tensor:
        """
        Shapes:
            k, q, v - (batch, heads, seq_len, width)
            attn_mask - (batch, seq_len)
        """

        if attn_mask.dim() != 2:
            raise ValueError(
                "The attention mask must be a 2D-tensor of shape [batch, seq_len]"
            )

        model_dim = k.shape[-1]
        attn_scores = q @ k.transpose(-2, -1)
        attn_scores /= math.sqrt(model_dim)

        # Replace tokens that we don't want to attend to with a large
        # negative value to zero them out during softmax normalization.
        batch, seq_len = attn_mask.shape
        attn_scores += attn_mask.logit_mask.view(batch, 1, 1, seq_len)

        attn_weights = attn_scores.softmax(dim=-1)
        attn_values = self.dropout(attn_weights @ v)

        return attn_values



from typing import Any, Dict, List, Optional, cast
import numpy
from pathlib import Path
from spacy import util
from spacy.cli._util import (
    debug_cli,
    import_code,
    parse_config_overrides,
    show_validation_error,
)
from spacy.schemas import ConfigSchemaTraining
from spacy.tokens import Doc
from spacy.util import registry, resolve_dot_names
from typer import Argument as Arg, Context, Option as Opt
from wasabi import Printer, msg

from ..tokenization.types import Tok2PiecesModelT
from ..pipeline.transformer import Transformer


@debug_cli.command(
    "pieces",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def debug_pieces_cli(
    # fmt: off
    ctx: Context,  # This is only used to read additional arguments
    config_path: Path = Arg(..., help="Path to config file", exists=True, allow_dash=True),
    code_path: Optional[Path] = Opt(None, "--code-path", "--code", "-c", help="Path to Python file with additional code (registered functions) to be imported"),
    transformer_name: Optional[str] = Opt(None, "--name", "-n", help="Name of the transformer pipe to gather piece statistics for (default: first transformer pipe)."),
    # fmt: on
):
    """
    Analyze word- or sentencepiece statistics.
    DOCS: https://spacy.io/api/cli#debug-pieces
    """
    overrides = parse_config_overrides(ctx.args)
    import_code(code_path)
    debug_pieces(
        config_path,
        config_overrides=overrides,
        transformer_name=transformer_name,
    )


def debug_pieces(
    config_path: Path,
    *,
    config_overrides: Dict[str, Any] = {},
    transformer_name: Optional[str] = None,
):
    msg = Printer()

    with show_validation_error(config_path):
        cfg = util.load_config(config_path, overrides=config_overrides)
        nlp = util.load_model_from_config(cfg, auto_fill=True)
        config = nlp.config.interpolate()
        T = registry.resolve(config["training"], schema=ConfigSchemaTraining)

    dot_names = [T["train_corpus"], T["dev_corpus"]]
    train_corpus, dev_corpus = resolve_dot_names(config, dot_names)

    nlp.initialize(lambda: train_corpus(nlp))

    if transformer_name is None:
        transformers = [
            pipe for _, pipe in nlp.pipeline if isinstance(pipe, Transformer)
        ]
        if not transformers:
            msg.fail("Pipeline does not contain transformer", exits=1)
        transformer_pipe = transformers[0]
    else:
        # We have to dance a bit around that MyPy cannot infer that we are
        # exiting if the invariants don't hold and that get_pipe returns
        # Callable.
        try:
            transformer_pipe_callable = nlp.get_pipe(transformer_name)
        except KeyError:
            transformer_pipe_callable = None
            msg.fail(
                f"Pipeline does not contain a pipe named '{transformer_name}'", exits=1
            )
        if not isinstance(transformer_pipe_callable, Transformer):
            msg.fail(f"Pipe named '{transformer_name}' is not a transformer", exits=1)
        transformer_pipe = cast(Transformer, transformer_pipe_callable)

    piece_encoder = transformer_pipe.model.get_ref("piece_encoder")
    msg.info(f"Found piece encoder: {piece_encoder.name}")

    train_docs = [eg.predicted for eg in train_corpus(nlp)]
    dev_docs = [eg.predicted for eg in dev_corpus(nlp)]

    msg.divider(f"Training corpus statistics")
    print_piece_stats(piece_encoder, train_docs)
    msg.divider(f"Development corpus statistics")
    print_piece_stats(piece_encoder, dev_docs)


def print_piece_stats(piece_encoder: Tok2PiecesModelT, docs: List[Doc]):
    docs_pieces = piece_encoder.predict(docs)

    lens = []
    for doc_pieces in docs_pieces:
        doc_piece_lens = doc_pieces.lengths
        lens.extend(doc_piece_lens.tolist())

    lens_xp = numpy.array(lens)

    msg.text(f"Median token length: {numpy.median(lens_xp)}")
    msg.text(f"Mean token length: {numpy.mean(lens_xp):.2f}")
    msg.text(f"Token length range: [{lens_xp.min()}, {lens_xp.max()}]")



import pytest
from cutlery import SentencePieceProcessor
from thinc.api import CupyOps, NumpyOps, Ragged, registry
from thinc.compat import has_cupy

from curated_transformers.models.architectures import (
    build_albert_transformer_model_v1,
    build_bert_transformer_model_v1,
    build_roberta_transformer_model_v1,
    build_xlmr_transformer_model_v1,
)
from curated_transformers.models.hf_loader import build_hf_transformer_encoder_loader_v1
from curated_transformers.models.output import TransformerModelOutput
from curated_transformers.models.with_strided_spans import build_with_strided_spans_v1
from curated_transformers.tokenization import (
    build_bert_wordpiece_encoder_v1,
    build_byte_bpe_encoder_v1,
    build_hf_piece_encoder_loader_v1,
    build_sentencepiece_encoder_v1,
    build_xlmr_sentencepiece_encoder_v1,
)
from curated_transformers._compat import (
    has_hf_transformers,
    has_huggingface_hub,
)


OPS = [NumpyOps()]
if has_cupy:
    OPS.append(CupyOps())


@pytest.fixture
def toy_model(test_dir):
    return SentencePieceProcessor.from_file(
        (str(test_dir / ".." / "tokenization" / "toy.model"))
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("stride,window", [(2, 4), (96, 128)])
@pytest.mark.parametrize("hf_model", [("xlm-roberta-base", 768, 250002)])
def test_xlmr_model(sample_docs, stride, window, hf_model):
    hf_model_name, hidden_width, vocab_size = hf_model
    with_spans = build_with_strided_spans_v1(stride=stride, window=window)
    model = build_xlmr_transformer_model_v1(
        piece_encoder=build_xlmr_sentencepiece_encoder_v1(),
        with_spans=with_spans,
        vocab_size=vocab_size,
        hidden_width=hidden_width,
    )
    model.get_ref("transformer").init = build_hf_transformer_encoder_loader_v1(
        name=hf_model_name
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name=hf_model_name
    )
    model.initialize(X=sample_docs)
    Y, backprop = model(sample_docs, is_train=False)
    assert isinstance(Y, TransformerModelOutput)
    num_ouputs = Y.num_outputs
    Y = Y.last_hidden_layer_states
    assert len(Y) == 2
    model.ops.xp.testing.assert_array_equal(Y[0].lengths, [1, 1, 1, 1, 1, 1, 2, 2])
    assert Y[0].dataXd.shape == (10, hidden_width)
    model.ops.xp.testing.assert_array_equal(Y[1].lengths, [1, 1, 1, 1, 2, 1, 2])
    assert Y[1].dataXd.shape == (9, hidden_width)

    # Backprop zeros to verify that backprop doesn't fail.
    dY = [
        [
            Ragged(
                model.ops.alloc2f(10, 768),
                lengths=model.ops.asarray1i([1, 1, 1, 1, 1, 1, 2, 2]),
            )
            for _ in range(num_ouputs)
        ],
        [
            Ragged(
                model.ops.alloc2f(9, 768),
                lengths=model.ops.asarray1i([1, 1, 1, 1, 2, 1, 2]),
            )
            for _ in range(num_ouputs)
        ],
    ]
    assert backprop(dY) == []


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("stride,window", [(2, 4), (96, 128)])
@pytest.mark.parametrize("hf_model", [("xlm-roberta-base", 768, 250002)])
def test_input_with_spaces(sample_docs_with_spaces, stride, window, hf_model):
    hidden_width = 768
    with_spans = build_with_strided_spans_v1(stride=stride, window=window)
    model = build_xlmr_transformer_model_v1(
        piece_encoder=build_xlmr_sentencepiece_encoder_v1(),
        with_spans=with_spans,
        vocab_size=250005,
        hidden_width=hidden_width,
        num_hidden_layers=1,
    )
    model.get_ref("piece_encoder").init = build_hf_piece_encoder_loader_v1(
        name="xlm-roberta-base"
    )
    model.initialize(X=sample_docs_with_spaces)
    Y, backprop = model(sample_docs_with_spaces, is_train=False)
    assert isinstance(Y, TransformerModelOutput)
    num_ouputs = Y.num_outputs
    Y = Y.last_hidden_layer_states
    assert len(Y) == 2
    model.ops.xp.testing.assert_array_equal(Y[0].lengths, [1, 1, 1, 1, 1, 1, 1, 2, 2])
    assert Y[0].dataXd.shape == (11, hidden_width)
    model.ops.xp.testing.assert_array_equal(
        Y[1].lengths, [1, 1, 1, 1, 1, 1, 2, 1, 2, 1]
    )
    assert Y[1].dataXd.shape == (12, hidden_width)

    # Backprop zeros to verify that backprop doesn't fail.
    dY = [
        [
            Ragged(
                model.ops.alloc2f(11, 768),
                lengths=model.ops.asarray1i([1, 1, 1, 1, 1, 1, 1, 2, 2]),
            )
            for _ in range(num_ouputs)
        ],
        [
            Ragged(
                model.ops.alloc2f(11, 768),
                lengths=model.ops.asarray1i([1, 1, 1, 1, 1, 2, 1, 2, 1]),
            )
            for _ in range(num_ouputs)
        ],
    ]
    assert backprop(dY) == []


@pytest.mark.slow
@pytest.mark.skipif(not has_huggingface_hub, reason="requires huggingface hub")
@pytest.mark.parametrize(
    "test_config",
    [
        (
            "albert-base-v2",
            build_albert_transformer_model_v1,
            build_sentencepiece_encoder_v1(),
            30000,
        ),
        (
            "bert-base-cased",
            build_bert_transformer_model_v1,
            build_bert_wordpiece_encoder_v1(),
            28996,
        ),
        (
            "roberta-base",
            build_roberta_transformer_model_v1,
            build_byte_bpe_encoder_v1(),
            50265,
        ),
    ],
)
def test_pytorch_checkpoint_loader(test_config):
    from huggingface_hub import hf_hub_download

    model_name, model_factory, piece_encoder, vocab_size = test_config

    checkpoint_path = hf_hub_download(repo_id=model_name, filename="pytorch_model.bin")
    with_spans = build_with_strided_spans_v1(stride=96, window=128)
    model = model_factory(
        piece_encoder=piece_encoder, vocab_size=vocab_size, with_spans=with_spans
    )
    model.get_ref("transformer").init = registry.model_loaders.get(
        "curated-transformers.PyTorchCheckpointLoader.v1"
    )(path=checkpoint_path)
    model.initialize()



import pytest
from spacy.util import registry as spacy_registry
from curated_transformers.util import registry


@pytest.mark.parametrize(
    "encoder_name",
    [
        "curated-transformers.BertWordpieceEncoder.v1",
        "curated-transformers.ByteBPEEncoder.v1",
        "curated-transformers.CamembertSentencepieceEncoder.v1",
        "curated-transformers.CharEncoder.v1",
        "curated-transformers.SentencepieceEncoder.v1",
        "curated-transformers.WordpieceEncoder.v1",
        "curated-transformers.XlmrSentencepieceEncoder.v1",
    ],
)
def test_encoder_from_registry(encoder_name):
    spacy_registry.architectures.get(encoder_name)()


@pytest.mark.parametrize(
    "loader_name",
    [
        "curated-transformers.ByteBPELoader.v1",
        "curated-transformers.CharEncoderLoader.v1",
        "curated-transformers.HFTransformerEncoderLoader.v1",
        "curated-transformers.HFPieceEncoderLoader.v1",
        "curated-transformers.PyTorchCheckpointLoader.v1",
        "curated-transformers.SentencepieceLoader.v1",
        "curated-transformers.WordpieceLoader.v1",
    ],
)
def test_encoder_loader_from_registry(loader_name):
    # Can't be constructed, since most loaders have mandatory arguments.
    registry.model_loaders.get(loader_name)



from itertools import islice
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
)

from spacy import Errors, Language, Vocab
from spacy.tokens import Doc
from spacy.pipeline import TrainablePipe
from spacy.training import Example, validate_examples, validate_get_examples
from spacy.util import minibatch
from thinc.api import Config, Optimizer, set_dropout_rate
from thinc.model import Model
from thinc.types import Ragged

from ..models.output import DocTransformerOutput, TransformerModelOutput
from ..models.listeners import TransformerListener

DEFAULT_CONFIG_STR = """
    [transformer]

    [transformer.model]
    @architectures = "curated-transformers.XlmrTransformer.v1"

    [transformer.model.piece_encoder]
    @architectures = "curated-transformers.XlmrSentencepieceEncoder.v1"

    [transformer.model.with_spans]
    @architectures = "curated-transformers.WithStridedSpans.v1"
"""


DEFAULT_CONFIG = Config().from_str(DEFAULT_CONFIG_STR)
DOC_EXT_ATTR = "trf_data"


@Language.factory(
    "curated_transformer",
    assigns=["doc._.trf_data"],
    default_config=DEFAULT_CONFIG["transformer"],
)
def make_transformer(
    nlp: Language,
    name: str,
    model: Model,
    *,
    frozen: bool = False,
    all_layer_outputs: bool = False,
) -> "Transformer":
    """Construct a Transformer component, which lets you plug a pre-trained
    transformer model into spaCy so you can use it in your pipeline. One or
    more subsequent spaCy components can use the transformer outputs as features
    in its model, with gradients backpropagated to the single shared weights.

    nlp (Language):
        The pipeline.
    name (str):
        The component instance name.
    model (Model):
        One of the supported pre-trained transformer models.
    frozen (bool):
        If `True`, the model's weights are frozen and no backpropagation is performed.
    all_layer_outputs (bool):
        If `True`, the model returns the outputs of all the layers. Otherwise, only the
        output of the last layer is returned. This must be set to `True` if any of the pipe's
        downstream listeners require the outputs of all transformer layers.
    """
    return Transformer(
        nlp.vocab,
        model,
        name=name,
        frozen=frozen,
        all_layer_outputs=all_layer_outputs,
    )


class Transformer(TrainablePipe):
    """spaCy pipeline component that provides access to a pre-trained transformer
    model from. Downstream components are connected to this pip using TransformerListener
    layers. This works similarly to spaCy's Transformer component and TransformerListener
    sublayer.

    The activations from the transformer are saved in the doc._.trf_data extension
    attribute.
    """

    def __init__(
        self,
        vocab: Vocab,
        model: Model,
        *,
        name: str = "transformer",
        frozen: bool = False,
        all_layer_outputs: bool = False,
    ) -> None:
        """
        vocab (Vocab):
            The shared vocabulary.
        model (Model):
            One of the supported pre-trained transformer models.
        name (str):
            The component instance name.
        frozen (bool):
            If `True`, the model's weights are frozen and no backpropagation is performed.
        all_layer_outputs (bool):
            If `True`, the model returns the outputs of all the layers. Otherwise, only the
            output of the last layer is returned. This must be set to `True` if any of the pipe's
            downstream listeners require the outputs of all transformer layers.
        """
        self.vocab = vocab
        self.model = model
        self.name = name
        self.listener_map: Dict[str, List[TransformerListener]] = {}
        self.cfg: Dict[str, Any] = {}

        _install_extensions()
        self.frozen = frozen
        self.all_layer_outputs = all_layer_outputs
        self._set_model_all_layer_outputs(all_layer_outputs)

    @property
    def listeners(self) -> List[TransformerListener]:
        """
        RETURNS (List[TransformerListener]):
            The listener models listening to this component. Usually internals.
        """
        return [m for c in self.listening_components for m in self.listener_map[c]]

    @property
    def listening_components(self) -> List[str]:
        """
        RETURNS (List[str]):
            The downstream components listening to this component. Usually internals.
        """
        return list(self.listener_map.keys())

    def add_listener(self, listener: TransformerListener, component_name: str) -> None:
        """Add a listener for a downstream component. Usually internals."""
        self.listener_map.setdefault(component_name, [])
        if listener not in self.listener_map[component_name]:
            self.listener_map[component_name].append(listener)

    def remove_listener(
        self, listener: TransformerListener, component_name: str
    ) -> bool:
        """Remove a listener for a downstream component. Usually internals.

        RETURNS (bool):
            `True` if successful, `False` otherwise.
        """
        if component_name in self.listener_map:
            if listener in self.listener_map[component_name]:
                self.listener_map[component_name].remove(listener)
                # If no listeners are left, remove entry
                if not self.listener_map[component_name]:
                    del self.listener_map[component_name]
                return True
        return False

    def find_listeners(self, component: Any) -> None:
        """Walk over a model of a processing component, looking for layers that
        are TransformerListener subclasses that have an upstream_name that matches
        this component. Listeners can also set their upstream_name attribute to
        the wildcard string '*' to match any `Transformer`.
        You're unlikely to ever need multiple `Transformer` components, so it's
        fine to leave your listeners upstream_name on '*'.
        """
        names = ("*", self.name)
        if isinstance(getattr(component, "model", None), Model):
            for node in component.model.walk():
                if (
                    isinstance(node, TransformerListener)
                    and node.upstream_name in names
                ):
                    self.add_listener(node, component.name)

    def pipe(self, stream: Iterable[Doc], *, batch_size: int = 128) -> Iterator[Doc]:
        """Apply the pipe to a stream of documents. This usually happens under
        the hood when the nlp object is called on a text and all components are
        applied to the Doc.

        stream (Iterable[Doc]):
            A stream of documents.
        batch_size (int):
            The number of documents to buffer.

        YIELDS (Doc):
            Processed documents in order.
        """
        _install_extensions()
        for batch in minibatch(stream, batch_size):
            preds = self.predict(batch)
            self.set_annotations(batch, preds)
            yield from batch

    def predict(self, docs: Iterable[Doc]) -> TransformerModelOutput:
        """Apply the pipeline's model to a batch of docs, without modifying them.

        docs (Iterable[Doc]):
            The documents to predict.

        RETURNS (TransformerModelOutput):
            The extracted features of each document.
        """
        _install_extensions()
        if not any(len(doc) for doc in docs):
            # Handle cases where there are no tokens in any docs.
            width = self.model.get_dim("nO")
            return TransformerModelOutput(
                outputs=[
                    [
                        Ragged(
                            data=self.model.ops.alloc2f(0, width),
                            lengths=self.model.ops.alloc1i(0),
                        )
                    ]
                    for _ in docs
                ],
                last_layer_only=True,
            )

        # To ensure that the model's internal state is always consistent with the pipe's.
        self._set_model_all_layer_outputs(self.all_layer_outputs)
        return self.model.predict(docs)

    def set_annotations(
        self, docs: Sequence[Doc], trf_output: TransformerModelOutput
    ) -> None:
        """Assign the extracted features to the Doc objects. By default, a
        DocTransformerOutput object is written to the doc._.trf_data attribute.

        docs (Iterable[Doc]):
            The documents to modify.
        trf_output (TransformerModelOutput):
            The outputs of the transformer model.
        """
        for doc, tokvecs in zip(docs, trf_output.all_outputs):
            doc._.trf_data = DocTransformerOutput(
                all_outputs=tokvecs, last_layer_only=trf_output.last_layer_only
            )

    def update(
        self,
        examples: Iterable[Example],
        *,
        drop: float = 0.0,
        sgd: Optional[Optimizer] = None,
        losses: Optional[Dict[str, float]] = None,
    ) -> Dict[str, float]:
        """Prepare for an update to the transformer.

        Like the `Tok2Vec` component, the `Transformer` component is unusual
        in that it does not receive "gold standard" annotations to calculate
        a weight update. The optimal output of the transformer data is unknown;
        it's a hidden layer inside the network that is updated by backpropagating
        from output layers.

        The `Transformer` component therefore does not perform a weight update
        during its own `update` method. Instead, it runs its transformer model
        and communicates the output and the backpropagation callback to any
        downstream components that have been connected to it via the
        TransformerListener sublayer. If there are multiple listeners, the last
        layer will actually backprop to the transformer and call the optimizer,
        while the others simply increment the gradients.

        examples (Iterable[Example]):
            A batch of Example objects. Only the `predicted` doc object is used,
            the reference doc is ignored.
        drop (float):
            The dropout rate.
        sgd (thinc.api.Optimizer):
            The optimizer.
        losses (Dict[str, float]):
            Optional record of the loss during training. Updated using the component
            name as the key.

        RETURNS (Dict[str, float]):
            The updated losses dictionary.
        """
        if losses is None:
            losses = {}
        validate_examples(examples, "Transformer.update")
        docs = [eg.predicted for eg in examples]
        set_dropout_rate(self.model, drop)
        losses.setdefault(self.name, 0.0)

        # To ensure that the model's internal state is always consistent with the pipe's.
        self._set_model_all_layer_outputs(self.all_layer_outputs)

        outputs, accum_func, backprop_func = self._create_backprops(
            docs, losses, sgd=sgd
        )

        batch_id = TransformerListener.get_batch_id(docs)
        for listener in self.listeners[:-1]:
            listener.receive(batch_id, outputs, accum_func)
        if self.listeners:
            self.listeners[-1].receive(batch_id, outputs, backprop_func)
        return losses

    def get_loss(self, examples: Iterable[Example], scores: Any) -> None:
        """A noop function, for compatibility with the Pipe API. See the `update`
        method for an explanation of the loss mechanics of the component.
        """
        pass

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        *,
        nlp: Optional[Language] = None,
        encoder_loader: Optional[Callable] = None,
        piecer_loader: Optional[Callable] = None,
    ):
        """Initialize the pipe for training, using data examples if available.

        get_examples (Callable[[], Iterable[Example]]):
            Optional function that returns gold-standard Example objects.
        nlp (Language):
            The current nlp object.
        encoder_loader (Optional[Callable]):
            Initialization callback for the transformer model.
        piece_loader (Optional[Callable]):
            Initialization callback for the input piece encoder.
        """
        validate_get_examples(get_examples, "Transformer.initialize")

        if encoder_loader:
            self.model.get_ref("transformer").init = encoder_loader  # type: ignore
        if piecer_loader:
            self.model.get_ref("piece_encoder").init = piecer_loader  # type: ignore

        doc_sample = []
        for example in islice(get_examples(), 10):
            doc_sample.append(example.x)
        assert doc_sample, Errors.E923.format(name=self.name)
        self.model.initialize(X=doc_sample)

    def add_label(self, label: Any):
        raise NotImplementedError

    def finish_update(self, sgd: Optimizer) -> None:
        """Update parameters using the current parameter gradients.
        The Optimizer instance contains the functionality to perform
        the stochastic gradient descent.

        This method is a noop when the pipe is frozen.

        sgd (thinc.api.Optimizer): The optimizer.

        DOCS: https://spacy.io/api/pipe#finish_update
        """
        if not self.frozen:
            self.model.finish_update(sgd)

    def _create_backprops(
        self,
        docs: Iterable[Doc],
        losses: Dict[str, float],
        *,
        sgd: Optional[Optimizer] = None,
    ) -> Tuple[TransformerModelOutput, Callable, Callable]:
        ops = self.model.ops
        # Accumulate the pipe's loss on the GPU at first.
        cum_loss = ops.xp.full(1, losses[self.name])

        if self.frozen:
            # Ensures that the inner torch model is executed in a `no_grad` context.
            outputs = self.model.predict(docs)
            bp_outputs = None
        else:
            outputs, bp_outputs = self.model.begin_update(docs)

        # We'll use this as a buffer for loss accumulation over individual gradients.
        d_outputs = [
            [Ragged(ops.alloc_f(t2v.dataXd.shape), t2v.lengths) for t2v in doc_layers]
            for doc_layers in outputs.all_outputs
        ]

        # The loss of the transformer is calculated in a different manner than that of the
        # tok2vec component. Instead of updating the loss value with each downstream component's
        # gradient, we accumulate the gradients and and update it just once. This reduces the overhead
        # of launching the associated kernels on the GPU as this operation would otherwise be
        # performed on a per-document, per-layer basis. The resultant loss value, while potentially
        # (slightly) different, still communicates the same information as before w.r.t the training
        # progress: how large were the gradients of the downstream components in this step compared
        # to the previous steps.

        def accumulate_gradient(
            one_d_outputs: List[List[Ragged]], outputs_to_backprop: Tuple[int, ...]
        ) -> None:
            """Accumulate transformer loss and gradient. This is passed as a callback
            to all but the last listener. Only the last one does the backprop.

            `outputs_to_backprop` is a tuple of indices indicating to which layers/outputs
            the gradients are to be propagated.
            """
            nonlocal d_outputs
            for i in range(len(one_d_outputs)):
                for j in outputs_to_backprop:
                    d_outputs[i][j].data += one_d_outputs[i][j].data

        def update_loss():
            """Reduce the gradient buffer and update the losses dict."""
            nonlocal cum_loss
            for i in range(len(d_outputs)):
                for j in range(len(d_outputs[i])):
                    cum_loss += (d_outputs[i][j].data ** 2).sum()
            losses[self.name] = float(cum_loss)

        def backprop(
            one_d_outputs: List[List[Ragged]], outputs_to_backprop: Tuple[int, ...]
        ) -> Any:
            """Callback to actually do the backprop. Passed to last listener."""
            nonlocal d_outputs
            assert bp_outputs is not None
            accumulate_gradient(one_d_outputs, outputs_to_backprop=outputs_to_backprop)
            update_loss()
            d_docs = bp_outputs(d_outputs)
            if sgd is not None:
                self.finish_update(sgd)
            return d_docs

        def backprop_noop(
            one_d_outputs: List[List[Ragged]], outputs_to_backprop: Tuple[int, ...]
        ) -> Any:
            accumulate_gradient(one_d_outputs, outputs_to_backprop=outputs_to_backprop)
            update_loss()
            if sgd is not None:
                self.finish_update(sgd)
            return []

        return outputs, accumulate_gradient, backprop_noop if self.frozen else backprop

    def _set_model_all_layer_outputs(self, new_value: bool):
        self.model.get_ref("transformer").attrs["_all_layer_outputs"] = new_value


def _install_extensions() -> None:
    if not Doc.has_extension(DOC_EXT_ATTR):
        Doc.set_extension(DOC_EXT_ATTR, default=None)



from typing import Tuple
from functools import lru_cache

from thinc.api import Model, Ragged

from .types import (
    PieceAdapterBackpropT,
    PieceAdapterInOutT,
    PieceAdapterModelT,
)

_FAIRSEQ_OFFSET = 1
_CAMEMBERT_FAIRSEQ_OFFSET = 4
_FAIRSEQ_BOS = 0
_FAIRSEQ_EOS = 2
_FAIRSEQ_UNK = 3

_SPP_BOS = 1
_SPP_EOS = 2
_SPP_UNK = 0


def _update_to_fairseq(piece_id: int):
    if piece_id == _SPP_UNK:
        return _FAIRSEQ_UNK
    elif piece_id == _SPP_BOS:
        return _FAIRSEQ_BOS
    elif piece_id == _SPP_EOS:
        return _FAIRSEQ_EOS
    else:
        return piece_id + _FAIRSEQ_OFFSET


@lru_cache(maxsize=128)
def _update_to_fairseq_vectorized(xp):
    return xp.vectorize(_update_to_fairseq)


def build_xlmr_adapter() -> PieceAdapterModelT:
    """Align the original fairseq vocab used by pre-trained HF transformer
    models with the sentencepiece vocabulary."""

    return Model(
        "xlmr_adapter",
        forward=xlmr_adapter_forward,
    )


def xlmr_adapter_forward(
    model: Model, X: PieceAdapterInOutT, is_train: bool
) -> Tuple[PieceAdapterInOutT, PieceAdapterBackpropT]:
    update_to_fairseq = _update_to_fairseq_vectorized(model.ops.xp)
    X_xlmr = []
    for tokens_pieces in X:
        X_xlmr.append(
            Ragged(
                data=update_to_fairseq(tokens_pieces.dataXd),
                lengths=tokens_pieces.lengths,
            )
        )

    return X_xlmr, lambda dY: []


def _camembert_update_to_fairseq(piece_id: int) -> int:
    if piece_id == _SPP_UNK:
        return _FAIRSEQ_UNK
    else:
        return piece_id + _CAMEMBERT_FAIRSEQ_OFFSET


@lru_cache(maxsize=128)
def _camembert_update_to_fairseq_vectorized(xp):
    return xp.vectorize(_camembert_update_to_fairseq)


def build_camembert_adapter() -> PieceAdapterModelT:
    """Align the original fairseq vocab used by pre-trained Camembert
    HF transformer model with the sentencepiece vocabulary."""
    return Model(
        "camembert_adapter",
        forward=camembert_adapter_forward,
    )


def camembert_adapter_forward(
    model: Model, X: PieceAdapterInOutT, is_train: bool
) -> Tuple[PieceAdapterInOutT, PieceAdapterBackpropT]:
    update_to_fairseq = _camembert_update_to_fairseq_vectorized(model.ops.xp)
    X_xlmr = []
    for tokens_pieces in X:
        X_xlmr.append(
            Ragged(
                data=update_to_fairseq(tokens_pieces.dataXd),
                lengths=tokens_pieces.lengths,
            )
        )

    return X_xlmr, lambda dY: []



class ErrorsWithCodes(type):
    def __getattribute__(self, code):
        msg = super().__getattribute__(code)
        if code.startswith("__"):  # python system attributes like __class__
            return msg
        else:
            return "(curated-transformers) [{code}] {msg}".format(code=code, msg=msg)


# fmt: off

class Warnings(metaclass=ErrorsWithCodes):
    W001 = ("Skipping transfomer pipe with TorchScript model - Model "
            "is likely already quantized")


class Errors(metaclass=ErrorsWithCodes):
    E001 = ("Attempting to quantize a transformer pipe with a "
            "non-PyTorch model ('{model_name}'). Quantization "
            "is only supported by PyTorch-specific curated "
            "transformer models")
    E002 = ("The number of hidden layers ({num_hidden_layers}) in the "
            "ALBERT encoder must be divisable by number of hidden groups "
            "({num_hidden_groups})")
    E003 = ("The hidden width of the transformer ({hidden_width}) must be "
            "divisible by the number of self-attention heads ({num_heads})")
    E004 = ("The point-wise feed-forward network in the transformer only "
            "supports the following activation functions: {activation_funcs}")
    E007 = ("Attempting to load the weights of an unsupported Hugging "
            "Face `transformers` model ({unsupported_model}). Currently "
            "supported models: {supported_models}")
    E008 = ("The number of layers in the scalar weighting listener model ({num_layers_scalar_weight})"
            "needs to be the same as the number of hidden layers in the "
            "transformer ({num_layers_transformer}). You can use the config's "
            "interpolation functionality to link the former's `num_layers` "
            "parameter with the latter")
    E009 = ("At least one sequence in the transformer's input has a length "
            "of {seq_len}, which is larger than the model's maximum sequence "
            "length of {max_seq_len} tokens")
    E010 = ("Curated transformers do not currently support listener replacement")
    E011 = ("`{loader_name}` requires the Hugging Face `transformers` package to be installed")
    E012 = ("`{listener_name}` requires the upstream transformer pipe to output "
            "all hidden layer outputs. This can be enabled by setting the pipe's "
            "`all_layer_outputs` parameter to `True` in the pipeline config")
    E013 = ("The target pipe names for gradual transformer unfreezing contain both the "
            "wild-card operator ('*') and individual names. Use either of the two but not both")
    E014 = ("Model '{model_name}' received an unexpected input of type '{input_type}'. "
            "It can only wrap/be chained with models whose outputs are of type  "
            "`TransformerModelOutput` (in almost all cases, models of type `TorchTransformerModelT`)")
    E015 = ("Input passed to the `ScalarWeight` model do not have the same number "
            "layers. Distinct layer counts: {layer_counts}")
    E016 = ("Input passed to the `ScalarWeight` model do not have the same width. "
            "Distinct widths: {hidden_widths}")
    E017 = ("Span extractor stride ({stride}) must be within [window_size / 2, "
            "window_size] ([{half_window_size}, {window_size}])")
    E018 = ("Span extractor batch size must be greater than zero")
    E019 = ("Byte-BPE piece encoder vocabulary doesn't contain '{piece}' piece")
    E020 = ("Character piece encoder vocabulary is not available. Use a loader "
            "to initialize the encoder.")
    E021 = ("Attempting to use the `CharEncoderLoader` piece encoder loader with an "
            "incompatible model ('{model_name}'). It can only be used with the "
            "`CharEncoder` piece encoder")
    E022 = ("Attempting to load an unsupported Hugging Face tokenizer "
            "({unsupported_tokenizer}). Currently supported tokenizers: "
            "{supported_tokenizers}")
    E023 = ("Japanese BERT models currently only support character subword encoding")
    E024 = ("Attempting to initialize an incompatible piece encoder ('{model_name}') "
            "with the Hugging Face Japanese BERT tokenizer. It can only be used with the "
            "`CharEncoder` piece encoder")
    E025 = ("Attempting to perform gradual unfreezing of a non-transformer pipe "
            "('{pipe_name}'}. Only transformer pipes support this feature")
    E026 = ("Attempting to load the weights of a Hugging Face `transformers` model "
            "into an unsupported curated encoder ({unsupported_encoder}). Currently "
            "supported encoders: {supported_encoders}")

# fmt: on



from spacy.cli import app
from typer.testing import CliRunner

import curated_transformers.cli.quantize
import curated_transformers.cli.debug_pieces


def test_quantize():
    result = CliRunner().invoke(app, ["quantize-transformer", "--help"])
    assert result.exit_code == 0


def test_debug_pieces():
    result = CliRunner().invoke(app, ["debug", "pieces", "--help"])
    assert result.exit_code == 0



from typing import List, Optional, Tuple, Iterator, Any, cast
from dataclasses import dataclass
from collections import Counter

from spacy.tokens import Doc
from thinc.api import Model, Ragged

from .types import (
    WsTokenAdapterInT,
    WsTokenAdapterOutT,
    WsTokenAdapterBackpropT,
    WsTokenAdapterModelT,
)
from ..tokenization.types import Tok2PiecesInT


@dataclass
class TokenAlignment:
    # The piece offset of the token within the doc with whitespace tokens.
    ws_piece_offset: int

    # The piece offset of the token within the doc without whitespace tokens.
    # Whitespace tokens have None as their offset, since they are not
    # represented in piece sequences without whitespace.
    no_ws_piece_offset: Optional[int]

    # The length of this sequence in pieces.
    n_pieces: int

    @property
    def is_whitespace(self) -> bool:
        return self.no_ws_piece_offset is None


@dataclass
class Alignment:
    # Token alignments.
    tokens: List[TokenAlignment]

    # The length of the document in pieces with whitespace tokens.
    ws_n_pieces: int

    # The length of the document in pieces without whitespace tokens.
    no_ws_n_pieces: int

    @property
    def has_no_whitespace(self) -> bool:
        return self.ws_n_pieces == self.no_ws_n_pieces

    def __iter__(self) -> Iterator[TokenAlignment]:
        return iter(self.tokens)


def with_non_ws_tokens(
    layer: Model[Tok2PiecesInT, WsTokenAdapterOutT]
) -> WsTokenAdapterModelT:
    """Removes non-whitespace tokens from the input before
    passing it to the inner model."""
    return Model(
        "with_non_ws_tokens",
        with_non_ws_tokens_forward,
        init=with_non_ws_tokens_init,
        layers=[layer],
    )


def with_non_ws_tokens_forward(
    model: Model, X: WsTokenAdapterInT, is_train: bool
) -> Tuple[WsTokenAdapterOutT, WsTokenAdapterBackpropT]:
    inner: Model[Tok2PiecesInT, WsTokenAdapterOutT] = model.layers[0]
    tokens, ws_counts = _filter_tokens(X)
    Y_no_ws, backprop_no_ws = inner(tokens, is_train)

    # Note: we modify the model outputs in-place. Since we are wrapping the
    # model, there should be no other consumers. Not sure yet if the same
    # applies to the gradient (e.g. consider summing two encoders of the
    # same width downstream.)

    alignments = _create_alignments(model, Y_no_ws, ws_counts)
    _add_whitespace_tokens(model, Y_no_ws, alignments)

    def backprop(dY: List[List[Ragged]]) -> Any:
        _remove_whitespace_tokens(model, dY, alignments)
        backprop_no_ws(dY)
        return []

    return Y_no_ws, backprop


def with_non_ws_tokens_init(
    model: WsTokenAdapterModelT,
    X: Optional[WsTokenAdapterInT] = None,
    Y: Optional[WsTokenAdapterInT] = None,
) -> None:
    model.layers[0].initialize(X=_filter_tokens(X)[0] if X is not None else None, Y=Y)


def _create_alignments(
    model: Model, output: WsTokenAdapterOutT, ws_counts: List[Counter]
) -> List[Alignment]:
    """Create an alignment between whitespace and non-whitespace sequences."""
    alignments = []
    for doc_output, doc_ws_counts in zip(output.all_outputs, ws_counts):
        doc_alignments = []
        no_ws_offset = 0
        ws_offset = 0
        pieces_lens = model.ops.to_numpy(doc_output[0].lengths).tolist()
        for idx, pieces_len in enumerate(pieces_lens):
            # Whitespace tokens that preceded a token.
            n_ws = doc_ws_counts[idx]
            for i in range(ws_offset, ws_offset + n_ws):
                doc_alignments.append(
                    TokenAlignment(
                        ws_piece_offset=i, no_ws_piece_offset=None, n_pieces=1
                    )
                )
            ws_offset += n_ws

            # The token itself.
            doc_alignments.append(
                TokenAlignment(
                    n_pieces=pieces_len,
                    no_ws_piece_offset=no_ws_offset,
                    ws_piece_offset=ws_offset,
                )
            )

            no_ws_offset += pieces_len
            ws_offset += pieces_len

        # There can be spaces after the last non-whitespace token.
        n_ws = doc_ws_counts[len(pieces_lens)]
        for i in range(ws_offset, ws_offset + n_ws):
            doc_alignments.append(
                TokenAlignment(ws_piece_offset=i, no_ws_piece_offset=None, n_pieces=1)
            )
        ws_offset += n_ws

        # Add the doc alignment.
        alignments.append(
            Alignment(
                tokens=doc_alignments,
                no_ws_n_pieces=no_ws_offset,
                ws_n_pieces=ws_offset,
            )
        )

    return alignments


def _filter_tokens(docs: List[Doc]) -> Tuple[Tok2PiecesInT, List[Counter]]:
    """Filter out whitespace tokens. Returns the non-whitespace tokens
    and a mapping from the (non-whitespace) token offset to the number
    of whitespaces that preceded the token."""
    tokens = []
    ws_counts = []
    for doc in docs:
        doc_tokens = []
        doc_ws_counts: Counter = Counter()
        offset = 0
        for token in doc:
            if token.is_space:
                doc_ws_counts[offset] += 1
                continue
            doc_tokens.append(token)
            offset += 1
        tokens.append(doc_tokens)
        ws_counts.append(doc_ws_counts)

    return tokens, ws_counts


def _add_whitespace_tokens(
    model: Model, Y_no_ws: WsTokenAdapterOutT, alignments: List[Alignment]
):
    """Add stub representations for whitespace tokens."""
    for Y_doc, doc_alignment in zip(Y_no_ws.all_outputs, alignments):
        if doc_alignment.has_no_whitespace:
            continue

        hidden_width = Y_doc[0].dataXd.shape[1]
        for layer_idx, layer in enumerate(Y_doc):
            lengths = []
            new_layer = model.ops.alloc2f(doc_alignment.ws_n_pieces, hidden_width)

            for alignment in doc_alignment:
                if not alignment.is_whitespace:
                    assert alignment.no_ws_piece_offset is not None
                    new_layer[
                        alignment.ws_piece_offset : alignment.ws_piece_offset
                        + alignment.n_pieces,
                        :,
                    ] = layer.dataXd[
                        alignment.no_ws_piece_offset : alignment.no_ws_piece_offset
                        + alignment.n_pieces,
                        :,
                    ]
                lengths.append(alignment.n_pieces)

            Y_doc[layer_idx] = Ragged(new_layer, lengths=model.ops.asarray1i(lengths))


def _remove_whitespace_tokens(
    model: Model, dY: List[List[Ragged]], alignments: List[Alignment]
):
    """Remove representations for whitespace tokens."""
    for dY_doc, doc_alignment in zip(dY, alignments):
        if doc_alignment.has_no_whitespace:
            continue

        hidden_width = cast(Tuple[int, ...], dY_doc[0].dataXd.shape)[1]
        for layer_idx, layer in enumerate(dY_doc):
            new_layer = model.ops.alloc2f(doc_alignment.no_ws_n_pieces, hidden_width)
            lengths = []

            for alignment in doc_alignment:
                if alignment.is_whitespace:
                    continue

                assert alignment.no_ws_piece_offset is not None
                # Extra type ignore to accomodate two MyPy versions :(.
                new_layer[  # type: ignore
                    alignment.no_ws_piece_offset : alignment.no_ws_piece_offset
                    + alignment.n_pieces,  # type: ignore
                    :,
                ] = layer.dataXd[  # type: ignore
                    alignment.ws_piece_offset : alignment.ws_piece_offset
                    + alignment.n_pieces,
                    :,
                ]

                lengths.append(alignment.n_pieces)

            dY_doc[layer_idx] = Ragged(new_layer, lengths=model.ops.asarray1i(lengths))



from typing import Callable, Optional, Tuple
from pathlib import Path
from cutlery import SentencePieceProcessor
from thinc.api import Model, Ragged, chain, deserialize_attr, serialize_attr

from .sentencepiece_adapters import build_camembert_adapter, build_xlmr_adapter

from .types import (
    Tok2PiecesBackpropT,
    Tok2PiecesInT,
    Tok2PiecesModelT,
    Tok2PiecesOutT,
)


@serialize_attr.register(SentencePieceProcessor)
def serialize_sentencepiece_processor(
    _, value: SentencePieceProcessor, name: str, model
) -> bytes:
    return value.to_protobuf()


@deserialize_attr.register(SentencePieceProcessor)
def deserialize_my_custom_class(
    _, value: bytes, name: str, model
) -> SentencePieceProcessor:
    return SentencePieceProcessor.from_protobuf(value)


def build_camembert_sentencepiece_encoder_v1() -> Tok2PiecesModelT:
    """Construct a SentencePiece piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers with CamemBERT post-processing applied.

    This model must be separately initialized using an appropriate
    loader.
    """
    encoder = build_sentencepiece_encoder_v1()
    model = chain(encoder, build_camembert_adapter())
    model.set_ref("encoder", encoder)
    return model


def build_sentencepiece_encoder_v1() -> Tok2PiecesModelT:
    """Construct a SentencePiece piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers.

    This model must be separately initialized using an appropriate
    loader.
    """
    model: Tok2PiecesModelT = Model(
        "sentencepiece_encoder",
        forward=sentencepiece_encoder_forward,
        attrs={"sentencepiece_processor": SentencePieceProcessor()},
    )
    model.set_ref("encoder", model)
    return model


def build_xlmr_sentencepiece_encoder_v1() -> Tok2PiecesModelT:
    """Construct a SentencePiece piece encoder model that accepts a list
    of token sequences or documents and returns a corresponding list
    of piece identifiers with XLM-RoBERTa post-processing applied.

    This model must be separately initialized using an appropriate
    loader.
    """
    encoder = build_sentencepiece_encoder_v1()
    model = chain(encoder, build_xlmr_adapter())
    model.set_ref("encoder", encoder)
    return model


def sentencepiece_encoder_forward(
    model: Model, X: Tok2PiecesInT, is_train: bool
) -> Tuple[Tok2PiecesOutT, Tok2PiecesBackpropT]:
    spp: SentencePieceProcessor = model.attrs["sentencepiece_processor"]

    pieces = []
    for doc in X:
        # TODO: check whether as single bos/eos per doc is what we want.
        #   The issue is that we probably do not have sentence
        #   boundaries yet, since they are predicted by a pipe.
        doc_pieces = [spp.bos_id()]
        lens = [1]

        for token in doc:
            piece_ids = spp.encode_as_ids(token.text)
            doc_pieces.extend(piece_ids)
            lens.append(len(piece_ids))

        doc_pieces.append(spp.eos_id())
        lens.append(1)
        pieces.append(
            Ragged(
                model.ops.asarray1i(doc_pieces),
                model.ops.asarray1i(lens),
            )
        )

    return pieces, lambda dY: []


def build_sentencepiece_encoder_loader_v1(
    *, path: Path
) -> Callable[
    [Tok2PiecesModelT, Optional[Tok2PiecesInT], Optional[Tok2PiecesInT]],
    Tok2PiecesModelT,
]:
    """Construct a callback that initializes a SentencePiece piece encoder
    model.

    path (Path):
        Path to the serialized SentencePiece model.
    """

    def load(model, X=None, Y=None):
        model.attrs["sentencepiece_processor"] = SentencePieceProcessor.from_file(
            str(path)
        )
        return model

    return load



from typing import Any, Callable, Iterable, List, Union

from spacy.tokens.doc import Doc
from thinc.model import Model
from thinc.types import Floats2d, Ints1d, Ragged

from .output import TransformerModelOutput

PoolingModelT = Model[Ragged, Floats2d]

WithRaggedLayersInT = Union[Iterable[Doc], Iterable[Iterable[Ragged]]]
WithRaggedLayersOutT = List[List[Floats2d]]  # Doc -> Layer -> Representation
WithRaggedLayersModelT = Model[WithRaggedLayersInT, WithRaggedLayersOutT]

WithRaggedLastLayerInT = Union[Iterable[Doc], Iterable[Ragged]]
WithRaggedLastLayerOutT = List[Floats2d]  # Doc -> Last Layer Representation
WithRaggedLastLayerModelT = Model[WithRaggedLastLayerInT, WithRaggedLastLayerOutT]


WsTokenAdapterInT = List[Doc]
WsTokenAdapterOutT = TransformerModelOutput
WsTokenAdapterBackpropT = Callable[[List[List[Ragged]]], Any]
WsTokenAdapterModelT = Model[WsTokenAdapterInT, WsTokenAdapterOutT]

# In case of a single list, each element corresponds to a single document/span.
# For nested lists, each inner list corresponds to a single document/span.
RaggedInOutT = Union[List[Ragged], List[List[Ragged]]]
Floats2dInOutT = Union[List[Floats2d], List[List[Floats2d]]]

SpanExtractorInT = List[Ragged]
SpanExtractorOutT = TransformerModelOutput
SpanExtractorBackpropT = Callable[[RaggedInOutT], RaggedInOutT]
SpanExtractorModelT = Model[SpanExtractorInT, SpanExtractorOutT]

# Inner PyTorch transformer.
TorchTransformerInT = List[Ints1d]
TorchTransformerOutT = TransformerModelOutput
TorchTransformerModelT = Model[TorchTransformerInT, TorchTransformerOutT]

SentMarkerRemoverInOutT = TransformerModelOutput
SentMarkerRemoverBackpropT = Callable[[RaggedInOutT], RaggedInOutT]
SentMarkerRemoverModelT = Model[SentMarkerRemoverInOutT, SentMarkerRemoverInOutT]

# Wrapper that includes the various pre- and post-processing layers such
# as the piece encoder, adapter, etc.
TransformerInT = List[Doc]
TransformerOutT = TransformerModelOutput
TransformerBackpropT = Callable[[List[List[Floats2d]]], Any]
TransformerModelT = Model[TransformerInT, TransformerOutT]

ScalarWeightInT = List[List[Ragged]]  # Doc -> Layer -> Representation
ScalarWeightOutT = List[Ragged]  # Doc -> Weighted Representation
ScalarWeightModelT = Model[ScalarWeightInT, ScalarWeightOutT]



from typing import List, Tuple, cast
from thinc.api import Model, Ragged

from .output import TransformerModelOutput
from .types import (
    SentMarkerRemoverInOutT,
    SentMarkerRemoverBackpropT,
    SentMarkerRemoverModelT,
    RaggedInOutT,
)
from ..errors import Errors


def remove_bos_eos() -> SentMarkerRemoverModelT:
    """Strips the BOS/EOS piece identifiers from the transformer output."""
    return Model("remove_bos_eos", remove_bos_eos_forward)


def remove_bos_eos_forward(
    model: Model, X: SentMarkerRemoverInOutT, is_train: bool
) -> Tuple[SentMarkerRemoverInOutT, SentMarkerRemoverBackpropT]:
    if not isinstance(X, TransformerModelOutput):
        raise ValueError(Errors.E014.format(model_name=model.name, input_type=type(X)))

    X.all_outputs = [[Xr[1:-1] for Xr in inner] for inner in X.all_outputs]

    def backprop(dY: RaggedInOutT) -> RaggedInOutT:
        # Pass-through dY, but add zero gradient for the special bos/eos
        # tokens.

        def _apply_to_layer(input: List[Ragged], output: List[Ragged]):
            for dYr in input:
                dim0 = dYr.dataXd.shape[0] + 2

                data = model.ops.xp.empty((dim0,) + dYr.dataXd.shape[1:], dtype="f")
                data[[0, -1]] = 0.0
                data[1:-1] = dYr.dataXd

                lengths = model.ops.alloc1i(dYr.lengths.shape[0] + 2, zeros=False)
                lengths[[0, -1]] = 1
                lengths[1:-1] = dYr.lengths
                output.append(Ragged(data, lengths=lengths))

        def _apply_to_layers(input: List[List[Ragged]], output: List[List[Ragged]]):
            for inner_dY in input:
                inner_dX: List[Ragged] = []
                _apply_to_layer(inner_dY, inner_dX)
                output.append(inner_dX)

        if isinstance(dY[0], list):
            nested_list = cast(List[List[Ragged]], dY)
            dX_nested: List[List[Ragged]] = []
            _apply_to_layers(nested_list, dX_nested)
            return dX_nested
        else:
            flat_list = cast(List[Ragged], dY)
            dX_flat: List[Ragged] = []
            _apply_to_layer(flat_list, dX_flat)
            return dX_flat

    return X, backprop



from typing import Callable
from dataclasses import dataclass
import pytest
from thinc.api import get_torch_default_device
from torch.nn import Module

from curated_transformers._compat import has_hf_transformers, transformers
from curated_transformers.models.architectures import _pytorch_encoder
from curated_transformers.models.hf_loader import build_hf_transformer_encoder_loader_v1
from curated_transformers.models.pytorch.albert import AlbertEncoder
from curated_transformers.models.pytorch.albert.config import AlbertConfig
from curated_transformers.models.pytorch.attention import AttentionMask
from curated_transformers.models.pytorch.bert import BertConfig, BertEncoder
from curated_transformers.models.pytorch.roberta.config import RobertaConfig
from curated_transformers.models.pytorch.roberta.encoder import RobertaEncoder

from ..util import torch_assertclose


@dataclass
class ModelConfig:
    config: BertConfig
    encoder: Callable[[BertConfig], Module]
    hf_model_name: str


TEST_MODELS = [
    ModelConfig(AlbertConfig(vocab_size=30000), AlbertEncoder, "albert-base-v2"),
    ModelConfig(BertConfig(vocab_size=28996), BertEncoder, "bert-base-cased"),
    ModelConfig(RobertaConfig(), RobertaEncoder, "roberta-base"),
    ModelConfig(RobertaConfig(vocab_size=250002), RobertaEncoder, "xlm-roberta-base"),
]


def encoder_from_config(config: ModelConfig):
    encoder = config.encoder(config.config)
    model = _pytorch_encoder(encoder)
    model.init = build_hf_transformer_encoder_loader_v1(name=config.hf_model_name)
    return model


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("model_config", TEST_MODELS)
def test_hf_load_weights(model_config):
    model = encoder_from_config(model_config)
    assert model


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("model_config", TEST_MODELS)
def test_model_against_hf_transformers(model_config):
    torch_device = get_torch_default_device()

    model = encoder_from_config(model_config)
    model.initialize()
    encoder = model.shims[0]._model
    encoder.eval()
    hf_encoder = transformers.AutoModel.from_pretrained(model_config.hf_model_name).to(
        torch_device
    )
    hf_encoder.eval()

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained(
        model_config.hf_model_name
    )
    tokenization = hf_tokenizer(
        ["This is a test.", "Let's match outputs"], padding=True, return_tensors="pt"
    ).to(torch_device)
    X = tokenization["input_ids"]
    attention_mask = tokenization["attention_mask"]

    # Test with the tokenizer's attention mask
    Y_encoder = encoder(
        X, attention_mask=AttentionMask(bool_mask=attention_mask.bool())
    )
    Y_hf_encoder = hf_encoder(X, attention_mask=attention_mask)

    torch_assertclose(
        Y_encoder.last_hidden_layer_states,
        Y_hf_encoder.last_hidden_state,
    )

    # Try to infer the attention mask from padding.
    Y_encoder = encoder(X)
    torch_assertclose(
        Y_encoder.last_hidden_layer_states,
        Y_hf_encoder.last_hidden_state,
    )



from typing import List, Optional, TypeVar, Generic
from dataclasses import dataclass
import torch
from torch import Tensor

from thinc.types import Floats2d, Ragged

from ..errors import Errors

TrfOutputT = TypeVar("TrfOutputT", Floats2d, Ragged)


@torch.jit.script
class PyTorchTransformerOutput:
    """Padded output of the PyTorch Transformer encoders."""

    # The first element is the output of the embedding layer with shape [batch, seq, width].
    # The rest of the elements are the states of each encoder hidden layer respectively with shape [batch, seq, width].
    all_outputs: List[Tensor]

    def __init__(
        self, *, embedding_output: Tensor, layer_hidden_states: List[Tensor]
    ) -> None:
        self.all_outputs = [embedding_output]
        self.all_outputs.extend(layer_hidden_states)

    @property
    def embedding_layer(self) -> Tensor:
        return self.all_outputs[0]

    def hidden_layer_states(self, idx: int) -> Tensor:
        """'idx' must be in the range [0, num_hidden_layers)"""
        if 0 <= idx < len(self.all_outputs) - 1:
            return self.all_outputs[idx + 1]
        else:
            # This error needs to be inlined as due to torch.jit.script limitations.
            raise ValueError(
                "Attempting to select a transformer output tensor using an invalid "
                f"layer index ({idx}). Expected range: 0<= idx < {(len(self.all_outputs) - 1)}"
            )

    @property
    def last_hidden_layer_states(self) -> Tensor:
        return self.all_outputs[-1]

    @property
    def all_hidden_layer_states(self) -> List[Tensor]:
        return self.all_outputs[1:]


@dataclass
class TransformerModelOutput(Generic[TrfOutputT]):
    """Wrapper for PyTorchTransformerOutput consumed by downstream non-PyTorch components.
    Also acts as the accumulator for the outputs of subsequent models in the Transformer pipeline."""

    # Non-padded, un-stacked versions of the outputs.
    # The outer list tracks Docs and the inner list
    # tracks the embedding + hidden layer outputs of each Doc.
    #
    # The inner-most element is a Floats2d when returned by
    # the PyTorchWrapper transformer model, which are subsequently
    # converted to Ragged by the models that follow.
    all_outputs: List[List[TrfOutputT]]

    # Set to True if only the last hidden layer's outputs are preserved.
    last_layer_only: bool

    def __init__(
        self, *, outputs: List[List[TrfOutputT]], last_layer_only: bool
    ) -> None:
        self.all_outputs = outputs
        self.last_layer_only = last_layer_only

    @property
    def embedding_layers(self) -> List[TrfOutputT]:
        if self.last_layer_only:
            return []
        else:
            return [y[0] for y in self.all_outputs]

    @property
    def last_hidden_layer_states(self) -> List[TrfOutputT]:
        return [y[-1] for y in self.all_outputs]

    @property
    def all_hidden_layer_states(self) -> List[List[TrfOutputT]]:
        return [y[1:] for y in self.all_outputs]

    @property
    def num_outputs(self) -> int:
        return len(self.all_outputs[0])


@dataclass
class DocTransformerOutput:
    """Stored on Doc instances. Each Ragged element corresponds to a layer in
    original TransformerModelOutput, containing representations of piece identifiers."""

    all_outputs: List[Ragged]

    # Set to True if only the last hidden layer's outputs are preserved.
    last_layer_only: bool

    def __init__(self, *, all_outputs: List[Ragged], last_layer_only: bool) -> None:
        self.all_outputs = all_outputs
        self.last_layer_only = last_layer_only

    @property
    def embedding_layer(self) -> Optional[Ragged]:
        if self.last_layer_only:
            return None
        else:
            return self.all_outputs[0]

    @property
    def last_hidden_layer_state(self) -> Ragged:
        return self.all_outputs[-1]

    @property
    def all_hidden_layer_states(self) -> List[Ragged]:
        return self.all_outputs[1:]

    @property
    def num_outputs(self) -> int:
        return len(self.all_outputs)



from typing import Callable, List, Optional
from spacy.tokens import Doc

from .._compat import transformers, has_hf_transformers
from .pytorch.hf_util import convert_hf_pretrained_model_parameters
from .types import TorchTransformerModelT
from ..errors import Errors


def build_hf_transformer_encoder_loader_v1(
    *,
    name: str,
    revision: str = "main",
) -> Callable[
    [TorchTransformerModelT, Optional[List[Doc]], Optional[List[Doc]]],
    TorchTransformerModelT,
]:
    """Construct a callback that initializes a supported transformer
    model with weights from a corresponding HuggingFace model.

    name (str):
        Name of the HuggingFace model.
    revision (str):
        Name of the model revision/branch.
    """

    def load(model, X=None, Y=None):
        if not has_hf_transformers:
            raise ValueError(
                Errors.E011.format(loader_name="HFTransformerEncoderLoader")
            )

        encoder = model.shims[0]._model

        hf_model = transformers.AutoModel.from_pretrained(name, revision=revision)
        params = convert_hf_pretrained_model_parameters(hf_model)
        encoder.load_state_dict(params)

        return model

    return load



import torch
from torch.nn import Linear, Module
from torch import Tensor

from .. import GeluNew
from ..attention import AttentionMask, ScaledDotProductAttention
from .config import BertAttentionConfig, BertLayerConfig
from ....errors import Errors


# https://www.tensorflow.org/text/tutorials/transformer#multi-head_attention
class BertSelfAttention(Module):
    def __init__(self, config: BertAttentionConfig):
        super().__init__()

        self.model_dim = config.hidden_width
        self.num_heads = config.num_attention_heads
        if self.model_dim % self.num_heads != 0:
            raise ValueError(
                Errors.E003.format(
                    hidden_width=self.model_dim, num_heads=self.num_heads
                )
            )

        self.dims_per_head = self.model_dim // self.num_heads
        self.attention = ScaledDotProductAttention(dropout_prob=config.dropout_prob)
        self.input = Linear(self.model_dim, self.model_dim * 3)
        self.output = Linear(self.model_dim, self.model_dim)

    def _split_heads(self, x: Tensor) -> Tensor:
        """
        Shapes:
            x - (batch, seq_len, width)
            output - (batch, head, seq_len, width_per_head)
        """
        batch_size, seq_len, model_dim = x.size()
        return x.view(
            batch_size, seq_len, self.num_heads, self.dims_per_head
        ).transpose(1, 2)

    def _combine_heads(self, x: Tensor) -> Tensor:
        """
        Shapes:
            x - (batch, head, seq_len, width_per_head)
            output - (batch, seq_len, width)
        """
        batch_size, head, seq_len, model_dim = x.size()
        return (
            x.transpose(1, 2).contiguous().view(batch_size, seq_len, head * model_dim)
        )

    def forward(self, x: Tensor, attn_mask: AttentionMask) -> Tensor:
        """
        Shapes:
            x - (batch, seq_len, width)
            attn_mask - (batch, seq_len)
        """

        proj = self.input(x)
        q, k, v = proj.chunk(3, dim=-1)

        # (batch, head, seq_len, width_per_head)
        k = self._split_heads(k)
        q = self._split_heads(q)
        v = self._split_heads(v)

        # (batch, seq_len, width)
        attn = self._combine_heads(self.attention(k, q, v, attn_mask))
        out = self.output(attn)

        return out


class BertFeedForward(Module):
    def __init__(self, config: BertLayerConfig):
        super().__init__()

        self.intermediate = Linear(config.hidden_width, config.intermediate_width)
        self.output = Linear(config.intermediate_width, config.hidden_width)
        if config.hidden_act == "relu":
            self.activation = torch.nn.ReLU()  # type: ignore
        elif config.hidden_act == "gelu":
            self.activation = torch.nn.GELU()  # type: ignore
        elif config.hidden_act == "gelu_new":
            # Ideally, we would use torch.nn.GELU(approximate="tanh"). However,
            # the differences between that and the manual Torch implementation
            # are large enough to fail tests comparing output to HF
            # transformers.
            self.activation = GeluNew()  # type: ignore
        else:
            raise ValueError(
                Errors.E004.format(activation_funcs=("relu", "gelu", "gelu_new"))
            )

    def forward(self, x: Tensor) -> Tensor:
        """
        Shapes:
            x - (batch, seq_len, width)
        """
        out = self.intermediate(x)
        out = self.activation(out)
        out = self.output(out)
        return out


class BertEncoderLayer(Module):
    def __init__(
        self, layer_config: BertLayerConfig, attention_config: BertAttentionConfig
    ):
        super().__init__()

        self.mha = BertSelfAttention(attention_config)
        self.attn_output_layernorm = torch.nn.LayerNorm(
            layer_config.hidden_width, eps=layer_config.layer_norm_eps
        )
        self.attn_output_dropout = torch.nn.Dropout(p=layer_config.dropout_prob)
        self.ffn = BertFeedForward(layer_config)
        self.ffn_output_layernorm = torch.nn.LayerNorm(
            layer_config.hidden_width, eps=layer_config.layer_norm_eps
        )
        self.ffn_output_dropout = torch.nn.Dropout(p=layer_config.dropout_prob)

    def forward(self, x: Tensor, attn_mask: AttentionMask) -> Tensor:
        """
        Shapes:
            x - (batch, seq_len, width)
            attn_mask - (batch, seq_len)
        """
        attn_out = self.mha(x, attn_mask)
        attn_out = self.attn_output_dropout(attn_out)
        attn_out = self.attn_output_layernorm(x + attn_out)

        ffn_out = self.ffn(attn_out)
        ffn_out = self.ffn_output_dropout(ffn_out)
        ffn_out = self.ffn_output_layernorm(attn_out + ffn_out)

        return ffn_out



from typing import Callable, Optional
import json


from .._compat import has_hf_transformers, transformers
from .bbpe_encoder import ByteBPEProcessor
from .sentencepiece_encoder import SentencePieceProcessor
from .wordpiece_encoder import WordPieceProcessor
from .types import (
    Tok2PiecesInT,
    Tok2PiecesModelT,
)
from ..errors import Errors

if has_hf_transformers:
    SUPPORTED_TOKENIZERS = (
        transformers.BertTokenizerFast,
        transformers.RobertaTokenizerFast,
        transformers.XLMRobertaTokenizerFast,
        transformers.CamembertTokenizerFast,
        transformers.BertJapaneseTokenizer,
    )
else:
    SUPPORTED_TOKENIZERS = ()  # type: ignore


def build_hf_piece_encoder_loader_v1(
    *, name: str, revision: str = "main"
) -> Callable[
    [Tok2PiecesModelT, Optional[Tok2PiecesInT], Optional[Tok2PiecesInT]],
    Tok2PiecesModelT,
]:
    """Construct a callback that initializes a HuggingFace piece encoder
    model. Used in conjunction with the HuggingFace model loader.

    name (str):
        Name of the HuggingFace model.
    revision (str):
        Name of the model revision/branch.
    """

    def load(model, X=None, Y=None):
        if not has_hf_transformers:
            raise ValueError(Errors.E011.format(loader_name="HFPieceEncoderLoader"))

        tokenizer = transformers.AutoTokenizer.from_pretrained(name, revision=revision)
        return _convert_encoder(model, tokenizer)

    return load


def _convert_encoder(
    model: Tok2PiecesModelT, tokenizer: "transformers.PreTrainedTokenizerBase"
) -> Tok2PiecesModelT:
    if isinstance(tokenizer, transformers.BertTokenizerFast):
        return _convert_wordpiece_encoder(model, tokenizer)
    elif isinstance(tokenizer, transformers.RobertaTokenizerFast):
        return _convert_byte_bpe_encoder(model, tokenizer)
    elif isinstance(
        tokenizer,
        (transformers.XLMRobertaTokenizerFast, transformers.CamembertTokenizerFast),
    ):
        return _convert_sentencepiece_encoder(model, tokenizer)  # type: ignore
    elif isinstance(tokenizer, transformers.BertJapaneseTokenizer):
        return _convert_bert_japanese_encoder(model, tokenizer)
    else:
        raise ValueError(
            Errors.E022.format(
                unsupported_tokenizer=type(tokenizer),
                supported_tokenizers=SUPPORTED_TOKENIZERS,
            )
        )


def _convert_byte_bpe_encoder(
    model: Tok2PiecesModelT,
    tokenizer: "transformers.RobertaTokenizerFast",
) -> Tok2PiecesModelT:
    # Seems like we cannot get the vocab file name for a RoBERTa vocabulary? And
    # neither the merges from the fast tokenizer. So we'll get them from the
    # JSON serialization.
    serialized = tokenizer.backend_tokenizer.to_str(True)  # type: ignore
    deserialized = json.loads(serialized)
    vocab_merges = deserialized["model"]
    merges = [tuple(merge.split(" ")) for merge in vocab_merges["merges"]]
    model.attrs["byte_bpe_processor"] = ByteBPEProcessor(vocab_merges["vocab"], merges)
    model.attrs["bos_piece"] = tokenizer.bos_token  # type: ignore
    model.attrs["eos_piece"] = tokenizer.eos_token  # type: ignore
    model.attrs["unk_piece"] = tokenizer.unk_token  # type: ignore

    return model


def _convert_sentencepiece_encoder(
    model: Tok2PiecesModelT,
    tokenizer: "transformers.RobertaTokenizerFast",
) -> Tok2PiecesModelT:
    model.get_ref("encoder").attrs[
        "sentencepiece_processor"
    ] = SentencePieceProcessor.from_file(
        tokenizer.vocab_file  # type: ignore
    )
    return model


def _convert_wordpiece_encoder(
    model: Tok2PiecesModelT, tokenizer: "transformers.BertTokenizerFast"
) -> Tok2PiecesModelT:
    # Seems like we cannot get the vocab file name for a BERT vocabulary? So,
    # instead, copy the vocabulary.
    vocab = [None] * tokenizer.vocab_size  # type: ignore
    for piece, idx in tokenizer.vocab.items():  # type: ignore
        vocab[idx] = piece

    strip_accents = tokenizer.backend_tokenizer.normalizer.strip_accents  # type: ignore
    lowercase = tokenizer.do_lower_case  # type: ignore
    model.attrs["wordpiece_processor"] = WordPieceProcessor(vocab)
    model.attrs["bos_piece"] = tokenizer.cls_token  # type: ignore
    model.attrs["eos_piece"] = tokenizer.sep_token  # type: ignore
    model.attrs["unk_piece"] = tokenizer.unk_token  # type: ignore
    model.attrs["lowercase"] = lowercase

    # Huggingface BERT also strips accents when lowercasing is enabled
    # and accent stripping is not defined.
    model.attrs["strip_accents"] = strip_accents or (
        strip_accents is not False and lowercase
    )

    return model


def _convert_bert_japanese_encoder(
    model: Tok2PiecesModelT, tokenizer: "transformers.BertJapaneseTokenizer"
) -> Tok2PiecesModelT:
    if not isinstance(
        tokenizer.subword_tokenizer,
        transformers.models.bert_japanese.CharacterTokenizer,
    ):
        raise ValueError(Errors.E023)
    if model.name != "char_encoder":
        raise ValueError(Errors.E024.format(model_name=model.name))

    model.attrs["bos_piece"] = tokenizer.cls_token
    model.attrs["eos_piece"] = tokenizer.sep_token
    model.attrs["unk_piece"] = tokenizer.unk_token
    model.attrs["normalize"] = (
        "NFKC" if tokenizer.subword_tokenizer.normalize_text else None
    )
    model.attrs["vocab"] = tokenizer.vocab.copy()

    return model



from typing import Optional
import torch
from torch.nn import Module
from torch import Tensor

from ....errors import Errors
from ..attention import AttentionMask
from ..bert.embeddings import BertEmbeddings
from ...output import PyTorchTransformerOutput
from .config import AlbertConfig
from .layer_group import AlbertLayerGroup


class AlbertEncoder(Module):
    def __init__(
        self,
        config: AlbertConfig,
    ):
        super().__init__()

        self.padding_idx = config.padding_idx
        self.max_seq_len = config.model_max_length
        self.num_hidden_layers = config.layer.num_hidden_layers
        num_hidden_groups = config.layer.num_hidden_groups

        if self.num_hidden_layers % num_hidden_groups != 0:
            raise ValueError(
                Errors.E002.format(
                    num_hidden_layers=self.num_hidden_layers,
                    num_hidden_groups=num_hidden_groups,
                )
            )

        self.embeddings = BertEmbeddings(config.embedding, config.layer)

        # Parameters are shared by groups of layers.
        self.groups = torch.nn.ModuleList(
            [
                AlbertLayerGroup(config.layer, config.attention)
                for _ in range(num_hidden_groups)
            ]
        )

    def _create_attention_mask(self, x: Tensor) -> AttentionMask:
        return AttentionMask(bool_mask=x.ne(self.padding_idx))

    def forward(
        self,
        input_ids: Tensor,
        attention_mask: Optional[AttentionMask] = None,
        token_type_ids: Optional[Tensor] = None,
    ) -> PyTorchTransformerOutput:
        """
        Shapes:
            input_ids, attention_mask, token_type_ids - (batch, seq_len)
        """
        if attention_mask is None:
            attention_mask = self._create_attention_mask(input_ids)

        embeddings = self.embeddings(input_ids, token_type_ids, None)
        layer_output = embeddings

        layers_per_group = self.num_hidden_layers // len(self.groups)

        layer_outputs = []
        for group in self.groups:
            for _ in range(layers_per_group):
                layer_output = group(layer_output, attn_mask=attention_mask)
                layer_outputs.append(layer_output)

        return PyTorchTransformerOutput(
            embedding_output=embeddings, layer_hidden_states=layer_outputs
        )



from typing import List, Union
import numpy
from spacy.tokens import Doc
from thinc.api import Model
from thinc.types import Ragged


from .types import WithRaggedLayersModelT, WithRaggedLastLayerModelT, PoolingModelT


def with_ragged_layers(
    pooling: PoolingModelT,
) -> WithRaggedLayersModelT:
    """Concatenate all Docs and layers into a Ragged for pooling."""
    return Model("with_ragged_layers", with_ragged_layers_forward, layers=[pooling])


def with_ragged_layers_forward(
    model: WithRaggedLayersModelT,
    X: Union[List[Doc], List[List[Ragged]]],
    is_train: bool,
):
    pooling: PoolingModelT = model.layers[0]
    xp = model.ops.xp

    datas = []
    lens = []
    doc_layer_counts = []
    doc_token_counts = []
    for X_doc in X:
        X_doc_layers = X_doc._.trf_data.all_outputs if isinstance(X_doc, Doc) else X_doc
        for X_layer in X_doc_layers:
            datas.append(X_layer.dataXd)
            lens.append(X_layer.lengths)
        doc_layer_counts.append(len(X_doc_layers))
        doc_token_counts.append(len(X_doc_layers[0].lengths))

    X_flat = Ragged(xp.concatenate(datas, axis=0), xp.concatenate(lens, axis=0))
    Y_pooled, pooling_backprop = pooling(X_flat, is_train)

    # This array is placed in CPU memory intentionally --- CuPy split (used
    # below) does not support indices in device memory.
    doc_layer_lens = numpy.asarray(
        [
            doc_layer * doc_len
            for doc_layer, doc_len in zip(doc_layer_counts, doc_token_counts)
        ]
    )
    Y = [
        xp.split(doc_array, doc_layer)
        for (doc_array, doc_layer) in zip(
            xp.split(Y_pooled, numpy.cumsum(doc_layer_lens)[:-1]),
            doc_layer_counts,
        )
    ]

    def backprop(dY):
        dY_pooled_flat = xp.concatenate(
            [dY_layer for dY_doc in dY for dY_layer in dY_doc]
        )
        dY_flat = pooling_backprop(dY_pooled_flat).dataXd

        dY = []
        for X_doc in X:
            dY_layer = []

            X_doc_layers = (
                X_doc._.trf_data.all_outputs if isinstance(X_doc, Doc) else X_doc
            )
            for X_layer in X_doc_layers:
                doc_unpooled_len = X_layer.dataXd.shape[0]
                dY_layer.append(Ragged(dY_flat[:doc_unpooled_len], X_layer.lengths))
                dY_flat = dY_flat[doc_unpooled_len:]

            dY.append(dY_layer)

        return dY

    return Y, backprop


def with_ragged_last_layer(
    pooling: PoolingModelT,
) -> WithRaggedLastLayerModelT:
    """Concatenate the last layers for all Docs into a Ragged for pooling."""
    return Model(
        "with_ragged_last_layer", with_ragged_last_layer_forward, layers=[pooling]
    )


def with_ragged_last_layer_forward(
    model: WithRaggedLastLayerModelT, X: Union[List[Doc], List[Ragged]], is_train: bool
):
    pooling: PoolingModelT = model.layers[0]
    xp = model.ops.xp

    datas = []
    lens = []
    doc_lens = []
    for X_doc in X:
        X_doc_layer = (
            X_doc._.trf_data.last_hidden_layer_state
            if isinstance(X_doc, Doc)
            else X_doc
        )
        datas.append(X_doc_layer.dataXd)
        lens.append(X_doc_layer.lengths)
        doc_lens.append(len(X_doc_layer.lengths))

    X_flat = Ragged(xp.concatenate(datas, axis=0), xp.concatenate(lens, axis=0))
    Y_pooled, pooling_backprop = pooling(X_flat, is_train)
    Y = xp.split(Y_pooled, numpy.cumsum(doc_lens)[:-1])

    def backprop(dY):
        dY_pooled_flat = xp.concatenate(dY)
        dY_flat = pooling_backprop(dY_pooled_flat).dataXd

        dY = []
        for X_doc in X:
            X_doc_layer = (
                X_doc._.trf_data.last_hidden_layer_states
                if isinstance(X_doc, Doc)
                else X_doc
            )
            doc_unpooled_len = X_doc_layer.dataXd.shape[0]
            dY.append(Ragged(dY_flat[:doc_unpooled_len], X_doc_layer.lengths))
            dY_flat = dY_flat[doc_unpooled_len:]

        return dY

    return Y, backprop
